
#include <cassert>

#include <cstddef>
using std::size_t;

#include <cmath>
using std::fabs;

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <vector>
using std::vector;

#include <numeric>
using std::accumulate;

#include <algorithm>
using std::set_difference;

#include "mtrand.hpp"
mtrand R;

#include "moments.hpp"

// global variables
vector<int> spin;
size_t L;
double T;
size_t attempts;

#ifdef USE_LINEAR_CHAIN
size_t site_count(size_t L) { return L; }
size_t right(size_t i) { return (i+1)%L; }
size_t left(size_t i) { return (L+i-1)%L; }
int neighbour_sum(size_t i) { return spin[right(i)] + spin[left(i)]; }
bool are_adjacent(size_t i, size_t j) { return j == right(i) or j == left(i); }
#endif

#ifdef USE_SQUARE_LATTICE
size_t site_count(size_t L) { return L*L; }
size_t right(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return (col+1)%L + row*L;
}
size_t left(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return (L+col-1)%L + row*L;
}
size_t up(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return col + ((row+1)%L)*L;
}
size_t down(size_t i) 
{ 
	const size_t row = i/L;
	const size_t col = i%L;
	return col + ((L+row-1)%L)*L;
}
int neighbour_sum(size_t i) 
{ 
	return spin[right(i)] + spin[left(i)]
	       + spin[up(i)] + spin[down(i)];
}
bool are_adjacent(size_t i, size_t j) { return j == right(i) or j == left(i) or j == up(i) or j == down(i); }
#endif

int recompute_energy(void)
{
	int sum = 0;
	for (size_t i = 0; i < spin.size(); ++i)
		sum += spin[i]*neighbour_sum(i);
	assert(sum%2 == 0);
	return -sum/2;
}

inline int recompute_magnetization(void)
{
	return accumulate(spin.begin(),spin.end(),0);
}

int E, dE, M, dM;
void initialize(size_t L_, double T_, bool ferro)
{
	L = L_;
	T = T_;
	spin.resize(site_count(L_));
	if (ferro)
		for (size_t i = 0; i < spin.size(); ++i)
			spin[i] = 1; // ferromagnetic initial state
	else
		for (size_t i = 0; i < spin.size(); ++i)
			spin[i] = ( R() < 0.5 ? 1 : -1); // random initial state
		
	E = recompute_energy();
	M = recompute_magnetization();
}

double w;
vector<size_t> sites_to_flip;

void attempt_single_flip(void)
{
	assert(sites_to_flip.empty());
	const size_t i = R.randIntExc(spin.size());
	sites_to_flip.push_back(i);
	dE = 2*spin[i]*neighbour_sum(i);
	dM = -2*spin[i];
}

void attempt_double_flip(void)
{
	assert(sites_to_flip.empty());
	const size_t i = R.randIntExc(spin.size());
	const size_t j = (1 + i + R.randIntExc(spin.size()-1))%spin.size();
	assert(i != j);
	sites_to_flip.push_back(i);
	sites_to_flip.push_back(j);
	dE = 2*( spin[i]*neighbour_sum(i) + spin[j]*neighbour_sum(j) );
	if (are_adjacent(i,j))
		dE -= 4*spin[i]*spin[j];
	dM = -2*(spin[i] + spin[j]);
}

void attempt_triple_flip(void)
{
	assert(sites_to_flip.empty());
	const size_t i = R.randIntExc(spin.size());
	const size_t j = (1 + i + R.randIntExc(spin.size()-1))%spin.size();
	assert(i != j);
	size_t k = i;
	while (k == i or k == j)
		k = R.randIntExc(spin.size());
	assert(k != i and k != j);
	sites_to_flip.push_back(i);
	sites_to_flip.push_back(j);
	sites_to_flip.push_back(k);
	dE = 2*( spin[i]*neighbour_sum(i) + spin[j]*neighbour_sum(j) + spin[k]*neighbour_sum(k));
	if (are_adjacent(i,j))
		dE -= 4*spin[i]*spin[j];
	if (are_adjacent(i,k))
		dE -= 4*spin[i]*spin[k];
	if (are_adjacent(j,k))
		dE -= 4*spin[j]*spin[k];
	dM = -2*(spin[i] + spin[j] + spin[k]);
}

void update(void)
{
	w = exp(-dE/T);
	attempts += sites_to_flip.size();
	if (w > 1.0 or w > R())
	{
		// accept
		while (!sites_to_flip.empty())
		{
			spin[ sites_to_flip.back() ] *= -1;
			sites_to_flip.pop_back();
		}
		assert(recompute_energy()-E == dE);
		E += dE;
		assert(recompute_magnetization()-M == dM);
		M += dM;		
	}
	else
	{
		// reject
		sites_to_flip.clear();		
	}
}

void cluster_update(void)
{
	// Wolff algorithm
	static vector<size_t> cluster_perim;
	cluster_perim.clear();
	const size_t i = R.randIntExc(spin.size());
	const int s = spin[i];
	cluster_perim.push_back(i);
	sites_to_flip.push_back(i);
	const double p = 1.0 - exp(-2.0/T);
	
	while (!cluster_perim.empty())
	{
		const size_t Ncl = cluster_perim.size();
		for (size_t n = 0; n < Ncl; ++n)
		{
			const size_t i = cluster_perim[n];
#if defined(USE_LINEAR_CHAIN) || defined(USE_SQUARE_LATTICE)			
			{
				const size_t j = right(i);
				if (spin[j] == s and R() < p) 
					cluster_perim.push_back(j);
			}
			{
				const size_t j = left(i);
				if (spin[j] == s and R() < p) 
					cluster_perim.push_back(j);
			}
#endif // USE_LINEAR_CHAIN or USE_SQUARE_LATTICE
#ifdef USE_SQUARE_LATTICE
			{
				const size_t j = up(i);
				if (spin[j] == s and R() < p) 
					cluster_perim.push_back(j);
			}
			{
				const size_t j = down(i);
				if (spin[j] == s and R() < p) 
					cluster_perim.push_back(j);
			}
#endif // USE_SQUARE_LATTICE			
		}
		sort(sites_to_flip.begin(),sites_to_flip.end());
		sort(cluster_perim.begin(),cluster_perim.end());
		// Remove duplicate elements from the cluster perimeter
		cluster_perim.resize(unique(cluster_perim.begin(),cluster_perim.end()) - cluster_perim.begin());
		// Remove elements that are already in the clusterf (i.e., in sites_to_flip)
		cluster_perim.resize(size_t(set_difference(cluster_perim.begin(),cluster_perim.end(),sites_to_flip.begin(),sites_to_flip.end(),cluster_perim.begin())-cluster_perim.begin()));
		for (size_t n = 0; n < cluster_perim.size(); ++n)
			sites_to_flip.push_back(cluster_perim[n]);
	}
	const int dM = -2*int(sites_to_flip.size())*s;
	attempts += sites_to_flip.size();
	// accept with probability 1
	while (!sites_to_flip.empty()) 
	{
		const size_t i = sites_to_flip.back();
		spin[i] *= -1;
		sites_to_flip.pop_back();
		//dE -= 2*spin[i]*neighbour_sum(i);
	}
	//cerr << recompute_energy()-E << "\t" << dE << endl;
	//assert(recompute_energy()-E == dE);
	E = recompute_energy();
	assert(recompute_magnetization()-M == dM);
	M += dM;		
}

void measure_magnetization(moments &O)
{
	//const int M = accumulate(spin.begin(),spin.end(),0);
	const double M1 = fabs(double(M));
	const double M2 = M1*M1;
	const double M4 = M2*M2;
	O.first += M1;
	O.second += M2;
	O.fourth += M4;
}

void measure_energy(moments &O)
{
	const double E1 = double(E);
	const double E2 = E1*E1;
	const double E4 = E2*E2;
	O.first += E1;
	O.second += E2;
	O.fourth += E4;
}
